@@ -15,7 +15,7 @@ from pytest import raises
 
 import h5netcdf
 from h5netcdf import legacyapi
-from h5netcdf.core import NOT_A_VARIABLE
+from h5netcdf.core import NOT_A_VARIABLE, CompatibilityError
 
 try:
     import h5pyd
@@ -68,6 +68,11 @@ def decode_vlen_strings(request):
         return {}
 
 
+@pytest.fixture(params=[netCDF4, legacyapi])
+def netcdf_write_module(request):
+    return request.param
+
+
 def get_hdf5_module(resource):
     """Return the correct h5py module based on the input resource."""
     if isinstance(resource, str) and resource.startswith(remote_h5):
@@ -129,6 +134,7 @@ def write_legacy_netcdf(tmp_netcdf, write_module):
     ds.createDimension("z", 6)
     ds.createDimension("empty", 0)
     ds.createDimension("string3", 3)
+    ds.createDimension("unlimited", None)
 
     v = ds.createVariable("foo", float, ("x", "y"), chunksizes=(4, 5), zlib=True)
     v[...] = 1
@@ -147,6 +153,9 @@ def write_legacy_netcdf(tmp_netcdf, write_module):
     v = ds.createVariable("intscalar", np.int64, (), zlib=6, fill_value=None)
     v[...] = 2
 
+    v = ds.createVariable("foo_unlimited", float, ("x", "unlimited"))
+    v[...] = 1
+
     with raises((h5netcdf.CompatibilityError, TypeError)):
         ds.createVariable("boolean", np.bool_, ("x"))
 
@@ -170,7 +179,7 @@ def write_h5netcdf(tmp_netcdf):
     ds = h5netcdf.File(tmp_netcdf, "w")
     ds.attrs["global"] = 42
     ds.attrs["other_attr"] = "yes"
-    ds.dimensions = {"x": 4, "y": 5, "z": 6, "empty": 0}
+    ds.dimensions = {"x": 4, "y": 5, "z": 6, "empty": 0, "unlimited": None}
 
     v = ds.create_variable(
         "foo", ("x", "y"), float, chunks=(4, 5), compression="gzip", shuffle=True
@@ -187,6 +196,9 @@ def write_h5netcdf(tmp_netcdf):
 
     v = ds.create_variable("intscalar", data=np.int64(2))
 
+    v = ds.create_variable("foo_unlimited", ("x", "unlimited"), float)
+    v[...] = 1
+
     with raises((h5netcdf.CompatibilityError, TypeError)):
         ds.create_variable("boolean", data=True)
 
@@ -221,10 +233,19 @@ def read_legacy_netcdf(tmp_netcdf, read_module, write_module):
     with pytest.raises(AttributeError):
         ds.does_not_exist
     assert set(ds.dimensions) == set(
-        ["x", "y", "z", "empty", "string3", "mismatched_dim"]
+        ["x", "y", "z", "empty", "string3", "mismatched_dim", "unlimited"]
     )
     assert set(ds.variables) == set(
-        ["foo", "y", "z", "intscalar", "scalar", "var_len_str", "mismatched_dim"]
+        [
+            "foo",
+            "y",
+            "z",
+            "intscalar",
+            "scalar",
+            "var_len_str",
+            "mismatched_dim",
+            "foo_unlimited",
+        ]
     )
     assert set(ds.groups) == set(["subgroup"])
     assert ds.parent is None
@@ -317,10 +338,19 @@ def read_h5netcdf(tmp_netcdf, write_module, decode_vlen_strings):
         # skip for now: https://github.com/Unidata/netcdf4-python/issues/388
         assert ds.attrs["other_attr"] == "yes"
     assert set(ds.dimensions) == set(
-        ["x", "y", "z", "empty", "string3", "mismatched_dim"]
+        ["x", "y", "z", "empty", "string3", "mismatched_dim", "unlimited"]
     )
     assert set(ds.variables) == set(
-        ["foo", "y", "z", "intscalar", "scalar", "var_len_str", "mismatched_dim"]
+        [
+            "foo",
+            "y",
+            "z",
+            "intscalar",
+            "scalar",
+            "var_len_str",
+            "mismatched_dim",
+            "foo_unlimited",
+        ]
     )
     assert set(ds.groups) == set(["subgroup"])
     assert ds.parent is None
@@ -402,7 +432,7 @@ def read_h5netcdf(tmp_netcdf, write_module, decode_vlen_strings):
     assert list(v.attrs) == []
 
     assert ds["/subgroup/y_var"].shape == (10,)
-    assert ds["/subgroup"].dimensions["y"] == 10
+    assert ds["/subgroup"].dimensions["y"].size == 10
 
     ds.close()
 
@@ -462,7 +492,7 @@ def test_fileobj(decode_vlen_strings):
 
 def test_repr(tmp_local_or_remote_netcdf):
     write_h5netcdf(tmp_local_or_remote_netcdf)
-    f = h5netcdf.File(tmp_local_or_remote_netcdf, "r")
+    f = h5netcdf.File(tmp_local_or_remote_netcdf, "a")
     assert "h5netcdf.File" in repr(f)
     assert "subgroup" in repr(f)
     assert "foo" in repr(f)
@@ -473,7 +503,7 @@ def test_repr(tmp_local_or_remote_netcdf):
 
     d = f.dimensions
     assert "h5netcdf.Dimensions" in repr(d)
-    assert "x=4" in repr(d)
+    assert "x=<h5netcdf.Dimension 'x': size 4>" in repr(d)
 
     g = f["subgroup"]
     assert "h5netcdf.Group" in repr(g)
@@ -485,9 +515,9 @@ def test_repr(tmp_local_or_remote_netcdf):
     assert "units" in repr(v)
 
     f.dimensions["temp"] = None
-    assert "temp: Unlimited (current: 0)" in repr(f)
+    assert "temp: <h5netcdf.Dimension 'temp': size 0 (unlimited)>" in repr(f)
     f.resize_dimension("temp", 5)
-    assert "temp: Unlimited (current: 5)" in repr(f)
+    assert "temp: <h5netcdf.Dimension 'temp': size 5 (unlimited)>" in repr(f)
 
     f.close()
 
@@ -531,7 +561,9 @@ def test_optional_netcdf4_attrs(tmp_local_or_remote_netcdf):
         f["foo"].dims[1].attach_scale(f["y"])
     with h5netcdf.File(tmp_local_or_remote_netcdf, "r") as ds:
         assert ds["foo"].dimensions == ("x", "y")
-        assert ds.dimensions == {"x": 5, "y": 10}
+        assert ds.dimensions.keys() == {"x", "y"}
+        assert ds.dimensions["x"].size == 5
+        assert ds.dimensions["y"].size == 10
         assert array_equal(ds["foo"], foo_data)
 
 
@@ -582,6 +614,13 @@ def test_mode_warning(tmp_local_or_remote_netcdf):
             pass
 
 
+def test_unlimited_chunk_warning(tmp_local_or_remote_netcdf):
+    with pytest.warns(FutureWarning):
+        with h5netcdf.File(tmp_local_or_remote_netcdf, "w") as ds:
+            ds.dimensions = {"x": None}
+            ds.create_variable("foo", ("x",), float)
+
+
 def create_invalid_netcdf_data():
     foo_data = np.arange(125).reshape(5, 5, 5)
     bar_data = np.arange(625).reshape(25, 5, 5)
@@ -625,21 +664,17 @@ def test_invalid_netcdf4(tmp_local_or_remote_netcdf):
                 fx.create_dataset(k, data=np.arange(v))
 
     with h5netcdf.File(tmp_local_or_remote_netcdf, "r", phony_dims="sort") as dsr:
-        i = len(grps) - 1
-        for grp in grps[::-1]:
+        for i, grp in enumerate(grps):
             var = dsr[grp].variables
             check_invalid_netcdf4(var, i)
-            i -= 1
 
     with h5netcdf.File(tmp_local_or_remote_netcdf, "r", phony_dims="access") as dsr:
-        for i, grp in enumerate(grps[::-1]):
-            print(dsr[grp])
+        for i, grp in enumerate(grps):
             var = dsr[grp].variables
             check_invalid_netcdf4(var, i)
 
     with netCDF4.Dataset(tmp_local_or_remote_netcdf, "r") as dsr:
         for i, grp in enumerate(grps):
-            print(dsr[grp])
             var = dsr[grp].variables
             check_invalid_netcdf4(var, i)
 
@@ -937,11 +972,15 @@ def test_creating_variables_with_unlimited_dimensions(tmp_local_or_remote_netcdf
             )
         assert e.value.args[0] == "Shape tuple is incompatible with data"
 
+        # Creating a coordinate variable
+        f.create_variable("x", dimensions=("x",), dtype=np.int64)
+
         # Resize data.
         assert f.variables["dummy"].shape == (0, 2)
         f.resize_dimension("x", 3)
         # This will also force a resize of the existing variables and it will
-        # be padded with zeros..
+        # be padded with zeros.
+        assert f.dimensions["x"].size == 3
         np.testing.assert_allclose(f.variables["dummy"], np.zeros((3, 2)))
 
         # Creating another variable with no data will now also take the shape
@@ -949,15 +988,27 @@ def test_creating_variables_with_unlimited_dimensions(tmp_local_or_remote_netcdf
         f.create_variable("dummy3", dimensions=("x", "y"), dtype=np.int64)
         assert f.variables["dummy3"].shape == (3, 2)
         assert f.variables["dummy3"]._h5ds.maxshape == (None, 2)
+        np.testing.assert_allclose(f.variables["dummy3"], np.zeros((3, 2)))
+
+        # Writing to a variable with an unlimited dimension raises
+        with pytest.raises(TypeError) as e:
+            f.variables["dummy3"][:] = np.ones((5, 2))
+        assert e.value.args[0] == "Can't broadcast (5, 2) -> (3, 2)"
+        assert f.variables["dummy3"].shape == (3, 2)
+        assert f.variables["dummy3"]._h5ds.maxshape == (None, 2)
+        assert f["x"].shape == (3,)
+        assert f.dimensions["x"].size == 3
+        np.testing.assert_allclose(f.variables["dummy3"], np.zeros((3, 2)))
 
     # Close and read again to also test correct parsing of unlimited
     # dimensions.
     with h5netcdf.File(tmp_local_or_remote_netcdf, "r") as f:
-        assert f.dimensions["x"] is None
+        assert f.dimensions["x"].isunlimited()
+        assert f.dimensions["x"].size == 3
         assert f._h5file["x"].maxshape == (None,)
         assert f._h5file["x"].shape == (3,)
 
-        assert f.dimensions["y"] == 2
+        assert f.dimensions["y"].size == 2
         assert f._h5file["y"].maxshape == (2,)
         assert f._h5file["y"].shape == (2,)
 
@@ -967,6 +1018,7 @@ def test_writing_to_an_unlimited_dimension(tmp_local_or_remote_netcdf):
         # Two dimensions, only one is unlimited.
         f.dimensions["x"] = None
         f.dimensions["y"] = 3
+        f.dimensions["z"] = None
 
         # Cannot create it without first resizing it.
         with pytest.raises(ValueError) as e:
@@ -979,6 +1031,7 @@ def test_writing_to_an_unlimited_dimension(tmp_local_or_remote_netcdf):
         f.create_variable("dummy1", dimensions=("x", "y"), dtype=np.int64)
         f.create_variable("dummy2", dimensions=("x", "y"), dtype=np.int64)
         f.create_variable("dummy3", dimensions=("x", "y"), dtype=np.int64)
+        f.create_variable("dummyX", dimensions=("x", "y", "z"), dtype=np.int64)
         g = f.create_group("test")
         g.create_variable("dummy4", dimensions=("y", "x", "x"), dtype=np.int64)
         g.create_variable("dummy5", dimensions=("y", "y"), dtype=np.int64)
@@ -986,20 +1039,22 @@ def test_writing_to_an_unlimited_dimension(tmp_local_or_remote_netcdf):
         assert f.variables["dummy1"].shape == (0, 3)
         assert f.variables["dummy2"].shape == (0, 3)
         assert f.variables["dummy3"].shape == (0, 3)
+        assert f.variables["dummyX"].shape == (0, 3, 0)
         assert g.variables["dummy4"].shape == (3, 0, 0)
         assert g.variables["dummy5"].shape == (3, 3)
+
+        # resize dimensions and all connected variables
         f.resize_dimension("x", 2)
         assert f.variables["dummy1"].shape == (2, 3)
         assert f.variables["dummy2"].shape == (2, 3)
         assert f.variables["dummy3"].shape == (2, 3)
+        assert f.variables["dummyX"].shape == (2, 3, 0)
         assert g.variables["dummy4"].shape == (3, 2, 2)
         assert g.variables["dummy5"].shape == (3, 3)
 
-        f.variables["dummy2"][:] = [[1, 2, 3], [5, 6, 7]]
-        np.testing.assert_allclose(f.variables["dummy2"], [[1, 2, 3], [5, 6, 7]])
-
-        f.variables["dummy3"][...] = [[1, 2, 3], [5, 6, 7]]
-        np.testing.assert_allclose(f.variables["dummy3"], [[1, 2, 3], [5, 6, 7]])
+        # broadcast writing
+        f.variables["dummy3"][...] = [[1, 2, 3]]
+        np.testing.assert_allclose(f.variables["dummy3"], [[1, 2, 3], [1, 2, 3]])
 
 
 def test_c_api_can_read_unlimited_dimensions(tmp_local_netcdf):
@@ -1045,21 +1100,26 @@ def test_reading_unlimited_dimensions_created_with_c_api(tmp_local_netcdf):
         # Assign something to trigger a resize.
         dummy1[:] = [[1, 2, 3], [4, 5, 6]]
 
+        # Create another variable with same dimensions
+        f.createVariable("dummy5", float, ("x", "y"))
+
     with h5netcdf.File(tmp_local_netcdf, "r") as f:
-        assert f.dimensions["x"] is None
-        assert f.dimensions["y"] == 3
-        assert f.dimensions["z"] is None
+        assert f.dimensions["x"].isunlimited()
+        assert f.dimensions["y"].size == 3
+        assert f.dimensions["z"].isunlimited()
 
         # This is parsed correctly due to h5netcdf's init trickery.
-        assert f._current_dim_sizes["x"] == 2
-        assert f._current_dim_sizes["y"] == 3
-        assert f._current_dim_sizes["z"] == 0
+        assert f.dimensions["x"].size == 2
+        assert f.dimensions["y"].size == 3
+        assert f.dimensions["z"].size == 0
 
         # But the actual data-set and arrays are not correct.
-        assert f["dummy1"].shape == (2, 3)
+        # assert f["dummy1"].shape == (2, 3)
         # XXX: This array has some data with dimension x - netcdf does not
         # appear to keep dimensions consistent.
-        assert f["dummy2"].shape == (3, 0, 0)
+        # With https://github.com/h5netcdf/h5netcdf/pull/103 h5netcdf will
+        # return a padded array
+        assert f["dummy2"].shape == (3, 2, 2)
         f.groups["test"]["dummy3"].shape == (3, 3)
         f.groups["test"]["dummy4"].shape == (0, 0)
 
@@ -1069,7 +1129,8 @@ def test_reading_unused_unlimited_dimension(tmp_local_or_remote_netcdf):
     with h5netcdf.File(tmp_local_or_remote_netcdf, "w") as f:
         f.dimensions = {"x": None}
         f.resize_dimension("x", 5)
-        assert f.dimensions == {"x": None}
+        assert f.dimensions["x"].isunlimited()
+        assert f.dimensions["x"].size == 5
 
 
 def test_reading_special_datatype_created_with_c_api(tmp_local_netcdf):
@@ -1092,7 +1153,10 @@ def test_nc4_non_coord(tmp_local_netcdf):
         f.create_variable("y", dimensions=("x",), dtype=np.int64)
 
     with h5netcdf.File(tmp_local_netcdf, "r") as f:
-        assert f.dimensions == {"x": None, "y": 2}
+        assert list(f.dimensions) == ["x", "y"]
+        assert f.dimensions["x"].size == 0
+        assert f.dimensions["x"].isunlimited()
+        assert f.dimensions["y"].size == 2
         assert list(f.variables) == ["y", "test"]
         assert list(f._h5group.keys()) == ["_nc4_non_coord_y", "test", "x", "y"]
 
@@ -1182,8 +1246,6 @@ def test_detach_scale(tmp_local_netcdf):
 
     with h5netcdf.File(tmp_local_netcdf, "a") as ds:
         ds.create_variable("test", dimensions=("x",), dtype=np.int64)
-        # this forces detach and re-creation
-        ds.create_variable("x", dimensions=("y",), dtype=np.int64)
 
     with h5netcdf.File(tmp_local_netcdf, "r") as ds:
         refs = ds._h5group["x"].attrs.get("REFERENCE_LIST", False)
@@ -1192,21 +1254,19 @@ def test_detach_scale(tmp_local_netcdf):
             assert dim == 0
             assert ds._root._h5file[ref].name == name
 
+    with h5netcdf.File(tmp_local_netcdf, "a") as ds:
+        ds.dimensions["x"]._detach_scale()
 
-def test_is_dimscale(tmp_local_netcdf):
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        refs = ds._h5group["x"].attrs.get("REFERENCE_LIST", False)
+        assert not refs
+
+
+def test_is_scale(tmp_local_netcdf):
     with legacyapi.Dataset(tmp_local_netcdf, "w") as ds:
         ds.createDimension("x", 10)
     with legacyapi.Dataset(tmp_local_netcdf, "r") as ds:
-        assert ds._is_dimscale("x")
-
-
-def test_delete_dim_scale(tmp_local_netcdf):
-    with legacyapi.Dataset(tmp_local_netcdf, "w") as ds:
-        ds.createDimension("x", 10)
-    with legacyapi.Dataset(tmp_local_netcdf, "r+") as ds:
-        assert ds._is_dimscale("x")
-        ds._delete_dim_scale("x")
-        assert not ds._is_dimscale("x")
+        assert ds.dimensions["x"]._isscale
 
 
 def test_get_dim_scale_refs(tmp_local_netcdf):
@@ -1215,7 +1275,7 @@ def test_get_dim_scale_refs(tmp_local_netcdf):
         ds.createVariable("test0", "i8", ("x",))
         ds.createVariable("test1", "i8", ("x",))
     with legacyapi.Dataset(tmp_local_netcdf, "r") as ds:
-        refs = ds._get_dim_scale_refs("x")
+        refs = ds.dimensions["x"]._scale_refs
         assert ds._h5file[refs[0][0]] == ds["test0"]._h5ds
         assert ds._h5file[refs[1][0]] == ds["test1"]._h5ds
 
@@ -1238,8 +1298,6 @@ def create_netcdf_dimensions(ds, idx):
     ship = g.createVariable("ship", "S1", ("ship", "ship_strlen"))
     sample = g.createVariable("sample", "i8", ("time", "sample"))
 
-    if isinstance(ds, legacyapi.Dataset):
-        g.resize_dimension("time", 10 + idx)
     time[:] = np.arange(10 + idx)
     data[:] = np.ones((3 + idx, 2 + idx, 10 + idx, 5 + idx)) * 12.0
     collide[...] = np.arange(5 + idx)
@@ -1300,18 +1358,18 @@ def check_netcdf_dimensions(tmp_netcdf, write_module, read_module):
                 "sample",
             }
             if read_module in [legacyapi, h5netcdf]:
-                assert g.dimensions["time"] is None
-                assert g._current_dim_sizes["time"] == 10 + i
-                assert g.dimensions["nvec"] is not None
-                assert g.dimensions["nvec"] == 5 + i
-                assert g.dimensions["sample"] is not None
-                assert g.dimensions["sample"] == 2 + i
-                assert g.dimensions["collide"] is not None
-                assert g.dimensions["collide"] == 7 + i
-                assert g.dimensions["ship"] is not None
-                assert g.dimensions["ship"] == 3 + i
-                assert g.dimensions["ship_strlen"] is not None
-                assert g.dimensions["ship_strlen"] == 10
+                assert g.dimensions["time"].isunlimited()
+                assert g.dimensions["time"].size == 10 + i
+                assert not g.dimensions["nvec"].isunlimited()
+                assert g.dimensions["nvec"].size == 5 + i
+                assert not g.dimensions["sample"].isunlimited()
+                assert g.dimensions["sample"].size == 2 + i
+                assert not g.dimensions["collide"].isunlimited()
+                assert g.dimensions["collide"].size == 7 + i
+                assert not g.dimensions["ship"].isunlimited()
+                assert g.dimensions["ship"].size == 3 + i
+                assert not g.dimensions["ship_strlen"].isunlimited()
+                assert g.dimensions["ship_strlen"].size == 10
             else:
                 assert g.dimensions["time"].isunlimited()
                 assert g.dimensions["time"].size == 10 + i
@@ -1338,10 +1396,7 @@ def check_netcdf_dimensions(tmp_netcdf, write_module, read_module):
             assert g.variables["data"].shape == (3 + i, 2 + i, 10 + i, 5 + i)
             assert g.variables["collide"].shape == (5 + i,)
             assert g.variables["non_collide"].shape == (5 + i,)
-            if write_module == netCDF4 and read_module in [legacyapi, h5netcdf]:
-                assert g.variables["sample"].shape == (2 + i, 2 + i)
-            else:
-                assert g.variables["sample"].shape == (10 + i, 2 + i)
+            assert g.variables["sample"].shape == (10 + i, 2 + i)
             assert g.variables["ship"].shape == (3 + i, 10)
 
 
@@ -1390,7 +1445,81 @@ def test_no_circular_references(tmp_local_netcdf):
 
     gc.collect()
     with h5netcdf.File(tmp_local_netcdf, "r") as ds:
-        assert len(gc.get_referrers(ds)) == 1
+        refs = gc.get_referrers(ds)
+        for ref in refs:
+            print(ref)
+        assert len(refs) == 1
+
+
+def test_expanded_variables_netcdf4(tmp_local_netcdf, netcdf_write_module):
+    with netcdf_write_module.Dataset(tmp_local_netcdf, "w") as ds:
+        f = ds.createGroup("test")
+        f.createDimension("x", None)
+        f.createDimension("y", 3)
+
+        dummy1 = f.createVariable("dummy1", float, ("x", "y"))
+        dummy2 = f.createVariable("dummy2", float, ("x", "y"))
+        dummy3 = f.createVariable("dummy3", float, ("x", "y"))
+        dummy4 = f.createVariable("dummy4", float, ("x", "y"))
+
+        dummy1[:] = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
+        dummy2[:] = [[1, 2, 3]]
+        dummy3[:] = [[1, 2, 3], [4, 5, 6]]
+
+        # don't mask, since h5netcdf doesn't do masking
+        if netcdf_write_module == netCDF4:
+            ds.set_auto_mask(False)
+
+        res1 = dummy1[:]
+        res2 = dummy2[:]
+        res3 = dummy3[:]
+        res4 = dummy4[:]
+
+    with netCDF4.Dataset(tmp_local_netcdf, "r") as ds:
+        # don't mask, since h5netcdf doesn't do masking
+        if netcdf_write_module == netCDF4:
+            ds.set_auto_mask(False)
+
+        f = ds["test"]
+
+        np.testing.assert_allclose(f.variables["dummy1"][:], res1)
+        assert f.variables["dummy1"].shape == (3, 3)
+        np.testing.assert_allclose(f.variables["dummy2"][:], res2)
+        assert f.variables["dummy2"].shape == (3, 3)
+        np.testing.assert_allclose(f.variables["dummy3"][:], res3)
+        assert f.variables["dummy3"].shape == (3, 3)
+        np.testing.assert_allclose(f.variables["dummy4"][:], res4)
+        assert f.variables["dummy4"].shape == (3, 3)
+
+    with legacyapi.Dataset(tmp_local_netcdf, "r") as ds:
+        f = ds["test"]
+        np.testing.assert_allclose(f.variables["dummy1"][:], res1)
+        assert f.variables["dummy1"].shape == (3, 3)
+        assert f.variables["dummy1"]._h5ds.shape == (3, 3)
+        np.testing.assert_allclose(f.variables["dummy2"][:], res2)
+        assert f.variables["dummy2"].shape == (3, 3)
+        assert f.variables["dummy2"]._h5ds.shape == (1, 3)
+        np.testing.assert_allclose(f.variables["dummy3"][:], res3)
+        assert f.variables["dummy3"].shape == (3, 3)
+        assert f.variables["dummy3"]._h5ds.shape == (2, 3)
+        np.testing.assert_allclose(f.variables["dummy4"][:], res4)
+        assert f.variables["dummy4"].shape == (3, 3)
+        assert f.variables["dummy4"]._h5ds.shape == (0, 3)
+
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        f = ds["test"]
+        np.testing.assert_allclose(f.variables["dummy1"][:], res1)
+        assert f.variables["dummy1"].shape == (3, 3)
+        assert f.variables["dummy1"]._h5ds.shape == (3, 3)
+        np.testing.assert_allclose(f.variables["dummy2"][:], res2)
+        assert f.variables["dummy2"].shape == (3, 3)
+        assert f.variables["dummy2"]._h5ds.shape == (1, 3)
+        np.testing.assert_allclose(f.variables["dummy3"][:], res3)
+        assert f.variables["dummy3"].shape == (3, 3)
+        assert f.variables["dummy3"]._h5ds.shape == (2, 3)
+        np.testing.assert_allclose(f.variables["dummy4"][:], res4)
+        assert f.variables["dummy4"].shape == (3, 3)
+        assert f.variables["dummy4"]._h5ds.shape == (0, 3)
 
 
 # https://github.com/h5netcdf/h5netcdf/issues/136
@@ -1436,15 +1565,6 @@ def test_track_order_false(tmp_local_netcdf):
         pass
 
 
-def test_monitor_netcdf4_requirement_for_track_order(tmp_local_netcdf):
-    # https://github.com/h5netcdf/h5netcdf/issues/130
-    with h5py.File(tmp_local_netcdf, "w", track_order=False):
-        pass
-    with pytest.raises(OSError):
-        with netCDF4.Dataset(tmp_local_netcdf, mode="a"):
-            pass
-
-
 # This should always work with the default file opening settings
 # https://github.com/h5netcdf/h5netcdf/issues/136#issuecomment-1017457067
 def test_more_than_7_attr_creation(tmp_local_netcdf):
@@ -1471,3 +1591,411 @@ def test_more_than_7_attr_creation_track_order(tmp_local_netcdf, track_order):
             for i in range(100):
                 h5file.attrs[f"key{i}"] = i
                 h5file.attrs[f"key{i}"] = 0
+
+
+def test_group_names(tmp_local_netcdf):
+    # https://github.com/h5netcdf/h5netcdf/issues/68
+    with netCDF4.Dataset(tmp_local_netcdf, mode="w") as ds:
+        for i in range(10):
+            ds = ds.createGroup(f"group{i:02d}")
+
+    with netCDF4.Dataset(tmp_local_netcdf, "r") as ds:
+        assert ds.name == "/"
+        name = ""
+        for i in range(10):
+            name = "/".join([name, f"group{i:02d}"])
+            assert ds[name].name == name.split("/")[-1]
+
+    with legacyapi.Dataset(tmp_local_netcdf, "r") as ds:
+        assert ds.name == "/"
+        name = ""
+        for i in range(10):
+            name = "/".join([name, f"group{i:02d}"])
+            assert ds[name].name == name.split("/")[-1]
+
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert ds.name == "/"
+        name = ""
+        for i in range(10):
+            name = "/".join([name, f"group{i:02d}"])
+            assert ds[name].name == name
+
+
+def test_legacyapi_endianess(tmp_local_netcdf):
+    # https://github.com/h5netcdf/h5netcdf/issues/15
+    big = legacyapi._check_return_dtype_endianess("big")
+    little = legacyapi._check_return_dtype_endianess("little")
+    native = legacyapi._check_return_dtype_endianess("native")
+
+    with legacyapi.Dataset(tmp_local_netcdf, "w") as ds:
+        ds.createDimension("x", 4)
+        # test creating variable using endian keyword argument
+        v = ds.createVariable("big", int, ("x"), endian="big")
+        v[...] = 65533
+        v = ds.createVariable("little", int, ("x"), endian="little")
+        v[...] = 65533
+        v = ds.createVariable("native", int, ("x"), endian="native")
+        v[...] = 65535
+
+    with h5py.File(tmp_local_netcdf, "r") as ds:
+        assert ds["big"].dtype.byteorder == big
+        assert ds["little"].dtype.byteorder == little
+        assert ds["native"].dtype.byteorder == native
+
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert ds["big"].dtype.byteorder == big
+        assert ds["little"].dtype.byteorder == little
+        assert ds["native"].dtype.byteorder == native
+
+    with legacyapi.Dataset(tmp_local_netcdf, "r") as ds:
+        assert ds["big"].dtype.byteorder == big
+        assert ds["little"].dtype.byteorder == little
+        assert ds["native"].dtype.byteorder == native
+
+    with netCDF4.Dataset(tmp_local_netcdf, "r") as ds:
+        assert ds["big"].dtype.byteorder == big
+        assert ds["little"].dtype.byteorder == little
+        assert ds["native"].dtype.byteorder == native
+
+
+def test_bool_slicing_length_one_dim(tmp_local_netcdf):
+    # see https://github.com/h5netcdf/h5netcdf/issues/23
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions = {"x": 1, "y": 2}
+        v = ds.create_variable("hello", ("x", "y"), "float")
+        v[:] = np.ones((1, 2))
+
+    bool_slice = np.array([1], dtype=bool)
+
+    # works for legacy API
+    with legacyapi.Dataset(tmp_local_netcdf, "a") as ds:
+        data = ds["hello"][bool_slice, :]
+        np.testing.assert_equal(data, np.ones((1, 2)))
+        ds["hello"][bool_slice, :] = np.zeros((1, 2))
+        data = ds["hello"][bool_slice, :]
+        np.testing.assert_equal(data, np.zeros((1, 2)))
+
+    # should raise for h5py >= 3.0.0
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        if version.parse(h5py.__version__) >= version.parse("3.0.0"):
+            error = "Indexing arrays must have integer dtypes"
+            with pytest.raises(TypeError) as e:
+                ds["hello"][bool_slice, :]
+            assert error == str(e.value)
+        else:
+            ds["hello"][bool_slice, :]
+
+
+def test_default_chunking(tmp_local_netcdf):
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions = {"x": 10, "y": 10, "z": 10, "t": None}
+
+        v = ds.create_variable(
+            "hello", ("x", "y", "z", "t"), "float", chunking_heuristic="h5py"
+        )
+        chunks_h5py = v.chunks
+
+        v = ds.create_variable(
+            "hello2", ("x", "y", "z", "t"), "float", chunking_heuristic=None
+        )
+        chunks_default = v.chunks
+
+        ds.resize_dimension("t", 4)
+        v = ds.create_variable(
+            "hello3", ("x", "y", "z", "t"), "float", chunking_heuristic="h5py"
+        )
+        chunks_resized = v.chunks
+
+    # cases above should be equivalent to a fixed dimension with appropriate size
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions = {"x": 10, "y": 10, "z": 10, "t": 1024}
+
+        v = ds.create_variable(
+            "hello",
+            ("x", "y", "z", "t"),
+            "float",
+            chunks=True,
+            chunking_heuristic="h5py",
+        )
+        chunks_true = v.chunks
+
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions = {"x": 10, "y": 10, "z": 10, "t": 4}
+
+        v = ds.create_variable(
+            "hello",
+            ("x", "y", "z", "t"),
+            "float",
+            chunks=True,
+            chunking_heuristic="h5py",
+        )
+        chunks_true_resized = v.chunks
+
+    assert chunks_h5py == chunks_true
+    assert chunks_default == chunks_true
+    assert chunks_resized == chunks_true_resized
+
+
+def test_h5netcdf_chunking(tmp_local_netcdf):
+    # produces much smaller chunks for unsized dimensions
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions = {"x": 10, "y": 10, "z": 10, "t": None}
+        v = ds.create_variable(
+            "hello", ("x", "y", "z", "t"), "float", chunking_heuristic="h5netcdf"
+        )
+        chunks_h5netcdf = v.chunks
+
+    assert chunks_h5netcdf == (10, 10, 10, 1)
+
+    # should produce chunks > 1 for small fixed dims
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions = {"x": 10, "t": None}
+        v = ds.create_variable(
+            "hello", ("x", "t"), "float", chunking_heuristic="h5netcdf"
+        )
+        chunks_h5netcdf = v.chunks
+
+    assert chunks_h5netcdf == (10, 128)
+
+    # resized unlimited dimensions should be treated like fixed dims
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions = {"x": 10, "y": 10, "z": 10, "t": None}
+        ds.resize_dimension("t", 10)
+        v = ds.create_variable(
+            "hello", ("x", "y", "z", "t"), "float", chunking_heuristic="h5netcdf"
+        )
+        chunks_h5netcdf = v.chunks
+
+    assert chunks_h5netcdf == (5, 5, 5, 10)
+
+
+def test_create_invalid_netcdf_catch_error(tmp_local_netcdf):
+    # see https://github.com/h5netcdf/h5netcdf/issues/138
+    with h5netcdf.File("test.nc", "w") as f:
+        try:
+            f.create_variable("test", ("x", "y"), data=np.ones((10, 10), dtype="bool"))
+        except CompatibilityError:
+            pass
+        assert repr(f.dimensions) == "<h5netcdf.Dimensions: >"
+
+
+def test_dimensions_in_parent_groups():
+    with netCDF4.Dataset("test_netcdf.nc", mode="w") as ds:
+        ds0 = ds
+        for i in range(10):
+            ds = ds.createGroup(f"group{i:02d}")
+        ds0.createDimension("x", 10)
+        ds0.createDimension("y", 20)
+        ds0["group00"].createVariable("test", float, ("x", "y"))
+        var = ds0["group00"].createVariable("x", float, ("x", "y"))
+        var[:] = np.ones((10, 20))
+
+    with legacyapi.Dataset("test_legacy.nc", mode="w") as ds:
+        ds0 = ds
+        for i in range(10):
+            ds = ds.createGroup(f"group{i:02d}")
+        ds0.createDimension("x", 10)
+        ds0.createDimension("y", 20)
+        ds0["group00"].createVariable("test", float, ("x", "y"))
+        var = ds0["group00"].createVariable("x", float, ("x", "y"))
+        var[:] = np.ones((10, 20))
+
+    with h5netcdf.File("test_netcdf.nc", mode="r") as ds0:
+        with h5netcdf.File("test_legacy.nc", mode="r") as ds1:
+            assert repr(ds0.dimensions["x"]) == repr(ds1.dimensions["x"])
+            assert repr(ds0.dimensions["y"]) == repr(ds1.dimensions["y"])
+            assert repr(ds0["group00"]) == repr(ds1["group00"])
+            assert repr(ds0["group00"]["test"]) == repr(ds1["group00"]["test"])
+            assert repr(ds0["group00"]["x"]) == repr(ds1["group00"]["x"])
+
+
+def test_array_attributes(tmp_local_netcdf):
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        dt = h5py.string_dtype("utf-8")
+        unicode = "unicodé"
+        ds.attrs["unicode"] = unicode
+        ds.attrs["unicode_0dim"] = np.array(unicode, dtype=dt)
+        ds.attrs["unicode_1dim"] = np.array([unicode], dtype=dt)
+        ds.attrs["unicode_arrary"] = np.array([unicode, "foobár"], dtype=dt)
+        ds.attrs["unicode_list"] = [unicode]
+
+        dt = h5py.string_dtype("ascii")
+        # if dtype is ascii it's irrelevant if the data is provided as bytes or string
+        ascii = "ascii"
+        ds.attrs["ascii"] = ascii
+        ds.attrs["ascii_0dim"] = np.array(ascii, dtype=dt)
+        ds.attrs["ascii_1dim"] = np.array([ascii], dtype=dt)
+        ds.attrs["ascii_array"] = np.array([ascii, "foobar"], dtype=dt)
+        ds.attrs["ascii_list"] = [ascii]
+
+        ascii = b"ascii"
+        ds.attrs["bytes"] = ascii
+        ds.attrs["bytes_0dim"] = np.array(ascii, dtype=dt)
+        ds.attrs["bytes_1dim"] = np.array([ascii], dtype=dt)
+        ds.attrs["bytes_array"] = np.array([ascii, b"foobar"], dtype=dt)
+        ds.attrs["bytes_list"] = [ascii]
+
+        dt = h5py.string_dtype("utf-8", 10)
+        # unicode needs to be encoded properly for fixed size string type
+        ds.attrs["unicode_fixed"] = np.array(unicode.encode("utf-8"), dtype=dt)
+        ds.attrs["unicode_fixed_0dim"] = np.array(unicode.encode("utf-8"), dtype=dt)
+        ds.attrs["unicode_fixed_1dim"] = np.array([unicode.encode("utf-8")], dtype=dt)
+        ds.attrs["unicode_fixed_arrary"] = np.array(
+            [unicode.encode("utf-8"), "foobár".encode("utf-8")], dtype=dt
+        )
+
+        dt = h5py.string_dtype("ascii", 10)
+        ascii = "ascii"
+        ds.attrs["ascii_fixed"] = np.array(ascii, dtype=dt)
+        ds.attrs["ascii_fixed_0dim"] = np.array(ascii, dtype=dt)
+        ds.attrs["ascii_fixed_1dim"] = np.array([ascii], dtype=dt)
+        ds.attrs["ascii_fixed_array"] = np.array([ascii, "foobar"], dtype=dt)
+
+        ascii = b"ascii"
+        ds.attrs["bytes_fixed"] = np.array(ascii, dtype=dt)
+        ds.attrs["bytes_fixed_0dim"] = np.array(ascii, dtype=dt)
+        ds.attrs["bytes_fixed_1dim"] = np.array([ascii], dtype=dt)
+        ds.attrs["bytes_fixed_array"] = np.array([ascii, b"foobar"], dtype=dt)
+
+        ds.attrs["int"] = 1
+        ds.attrs["intlist"] = [1]
+        ds.attrs["int_array"] = np.arange(10)
+
+    with h5netcdf.File(tmp_local_netcdf, mode="r") as ds:
+        assert ds.attrs["unicode"] == unicode
+        assert ds.attrs["unicode_0dim"] == unicode
+        assert ds.attrs["unicode_1dim"] == unicode
+        assert ds.attrs["unicode_arrary"] == [unicode, "foobár"]
+        assert ds.attrs["unicode_list"] == unicode
+
+        # bytes and strings are received as strings for h5py3
+        if version.parse(h5py.__version__) >= version.parse("3.0.0"):
+            ascii = "ascii"
+            foobar = "foobar"
+        # and bytes for h5py2
+        else:
+            ascii = b"ascii"
+            foobar = b"foobar"
+        assert ds.attrs["ascii"] == "ascii"
+        assert ds.attrs["ascii_0dim"] == ascii
+        assert ds.attrs["ascii_1dim"] == ascii
+        assert ds.attrs["ascii_array"] == [ascii, foobar]
+        # list is decoded for h5py2
+        assert ds.attrs["ascii_list"] == "ascii"
+
+        assert ds.attrs["bytes"] == ascii
+        assert ds.attrs["bytes_0dim"] == ascii
+        assert ds.attrs["bytes_1dim"] == ascii
+        assert ds.attrs["bytes_array"] == [ascii, foobar]
+        # list is decoded for h5py2
+        assert ds.attrs["bytes_list"] == "ascii"
+
+        assert ds.attrs["unicode_fixed"] == unicode
+        assert ds.attrs["unicode_fixed_0dim"] == unicode
+        assert ds.attrs["unicode_fixed_1dim"] == unicode
+        assert ds.attrs["unicode_fixed_arrary"] == [unicode, "foobár"]
+
+        ascii = "ascii"
+        assert ds.attrs["ascii_fixed"] == ascii
+        assert ds.attrs["ascii_fixed_0dim"] == ascii
+        assert ds.attrs["ascii_fixed_1dim"] == ascii
+        assert ds.attrs["ascii_fixed_array"] == [ascii, "foobar"]
+
+        assert ds.attrs["bytes_fixed"] == ascii
+        assert ds.attrs["bytes_fixed_0dim"] == ascii
+        assert ds.attrs["bytes_fixed_1dim"] == ascii
+        assert ds.attrs["bytes_fixed_array"] == [ascii, "foobar"]
+
+        assert ds.attrs["int"] == 1
+        assert ds.attrs["intlist"] == 1
+        np.testing.assert_equal(ds.attrs["int_array"], np.arange(10))
+
+    with legacyapi.Dataset(tmp_local_netcdf, mode="r") as ds:
+        assert ds.unicode == unicode
+        assert ds.unicode_0dim == unicode
+        assert ds.unicode_1dim == unicode
+        assert ds.unicode_arrary == [unicode, "foobár"]
+        assert ds.unicode_list == unicode
+
+        # bytes and strings are received as strings for h5py3
+        if version.parse(h5py.__version__) >= version.parse("3.0.0"):
+            ascii = "ascii"
+            foobar = "foobar"
+        # and bytes for h5py2
+        else:
+            ascii = b"ascii"
+            foobar = b"foobar"
+        assert ds.ascii == "ascii"
+        assert ds.ascii_0dim == ascii
+        assert ds.ascii_1dim == ascii
+        assert ds.ascii_array == [ascii, foobar]
+        # list is decoded for h5py2
+        assert ds.ascii_list == "ascii"
+
+        assert ds.bytes == ascii
+        assert ds.bytes_0dim == ascii
+        assert ds.bytes_1dim == ascii
+        assert ds.bytes_array == [ascii, foobar]
+        # list is decoded for h5py2
+        assert ds.bytes_list == "ascii"
+
+        assert ds.unicode_fixed == unicode
+        assert ds.unicode_fixed_0dim == unicode
+        assert ds.unicode_fixed_1dim == unicode
+        assert ds.unicode_fixed_arrary == [unicode, "foobár"]
+
+        ascii = "ascii"
+        assert ds.ascii_fixed == ascii
+        assert ds.ascii_fixed_0dim == ascii
+        assert ds.ascii_fixed_1dim == ascii
+        assert ds.ascii_fixed_array == [ascii, "foobar"]
+
+        assert ds.bytes_fixed == ascii
+        assert ds.bytes_fixed_0dim == ascii
+        assert ds.bytes_fixed_1dim == ascii
+        assert ds.bytes_fixed_array == [ascii, "foobar"]
+
+        assert ds.int == 1
+        assert ds.intlist == 1
+        np.testing.assert_equal(ds.int_array, np.arange(10))
+
+    with netCDF4.Dataset(tmp_local_netcdf, mode="r") as ds:
+        assert ds.unicode == unicode
+        assert ds.unicode_0dim == unicode
+        assert ds.unicode_1dim == unicode
+        assert ds.unicode_arrary == [unicode, "foobár"]
+        assert ds.unicode_list == unicode
+
+        ascii = "ascii"
+        assert ds.ascii == ascii
+        assert ds.ascii_0dim == ascii
+        assert ds.ascii_1dim == ascii
+        assert ds.ascii_array == [ascii, "foobar"]
+        assert ds.ascii_list == ascii
+
+        assert ds.bytes == ascii
+        assert ds.bytes_0dim == ascii
+        assert ds.bytes_1dim == ascii
+        assert ds.bytes_array == [ascii, "foobar"]
+        # writing/reading lists is broken with h5py2/netCDF4
+        if version.parse(h5py.__version__) >= version.parse("3.0.0"):
+            assert ds.bytes_list == ascii
+
+        assert ds.unicode_fixed == unicode
+        assert ds.unicode_fixed_0dim == unicode
+        assert ds.unicode_fixed_1dim == unicode
+        assert ds.unicode_fixed_arrary == [unicode, "foobár"]
+
+        assert ds.ascii_fixed == ascii
+        assert ds.ascii_fixed_0dim == ascii
+        assert ds.ascii_fixed_1dim == ascii
+        assert ds.ascii_fixed_array == [ascii, "foobar"]
+
+        assert ds.bytes_fixed == ascii
+        assert ds.bytes_fixed_0dim == ascii
+        assert ds.bytes_fixed_1dim == ascii
+        assert ds.bytes_fixed_array == [ascii, "foobar"]
+
+        assert ds.int == 1
+        assert ds.intlist == 1
+        np.testing.assert_equal(ds.int_array, np.arange(10))
